==> serenity-js-1.0.0.tgz: package/src/serenity/reporting/rehearsal_report.ts <==
import {
    Activity,
    ActivityFinished,
    ActivityStarts,
    DomainEvent,
    Outcome,
    Photo,
    PhotoAttempted,
    Scene,
    SceneFinished,
    SceneStarts,
} from '../../serenity/domain';
import { ReportExporter } from './report_exporter';

export class RehearsalReport {
    static from(events: Array<DomainEvent<any>>): RehearsalPeriod {
        let previousNode: ReportPeriod<Scene | Activity> = undefined;
        let currentNode:  ReportPeriod<Scene | Activity> = undefined;

        return events.reduce((fullReport, event) => {
            switch (event.constructor) {    // tslint:disable-line:switch-default   - ignore other events
                case SceneStarts:
                    currentNode = fullReport.append(new ScenePeriod(event));
                    break;
                case ActivityStarts:
                    currentNode = currentNode.append(new ActivityPeriod(event));
                    break;
                case SceneFinished:
                    currentNode = currentNode.concludeWith(event);
                    break;
                case ActivityFinished:
                    // The Photographer triggers a PhotoAttempted event after the ActivityFinished,
                    // that's why we need to cache both the previous and the current node
                    previousNode = currentNode;
                    currentNode  = currentNode.concludeWith(event);
                    break;
                case PhotoAttempted:
                    [previousNode, currentNode]
                        .filter(node => !! node && node.matches(event.value.activity))
                        .forEach(node => node.attach(event.value.photo));
                    break;
            }

            return fullReport;
        }, new RehearsalPeriod());
    }
}

export abstract class ReportPeriod<T> {
    public parent:    ReportPeriod<any>;
    public value:     T;

    public children:  Array<ReportPeriod<any>> = [];
    public outcome:   Outcome<T>;

    public startedAt:  number;
    protected finishedAt: number;

    private promisedPhotos: Array<PromiseLike<Photo>> = [];

    constructor(start: DomainEvent<T>) {
        this.value     = start.value;
        this.startedAt = start.timestamp;
    }

    abstract matches(finished: T): boolean;

    concludeWith(finished: DomainEvent<Outcome<any>>) {
        this.finishedAt = finished.timestamp;
        this.outcome    = finished.value;

        if (! this.parent) {
            return this;
        }

        return this.matches(finished.value.subject)
            ? this.parent
            : this.parent.concludeWith(finished);
    }

    attach(promisedPhoto: PromiseLike<Photo>) {
        this.promisedPhotos.push(promisedPhoto);
    }

    photos(): PromiseLike<Photo[]> {
        // todo: maybe use some constant rather than a blunt "undefined"?
        return Promise.all(this.promisedPhotos).then(photos => photos.filter(p => p !== undefined));
    }

    duration() {
        return this.finishedAt - this.startedAt;
    }

    append(child: ReportPeriod<Scene | Activity>): ReportPeriod<Scene | Activity> {
        child.parent = this;
        this.children.push(child);

        return child;
    }

    abstract exportedUsing<FORMAT>(exporter: ReportExporter<FORMAT>): PromiseLike<FORMAT>;
}

export class RehearsalPeriod extends ReportPeriod<Rehearsal> {
    matches(finished: Rehearsal): boolean {
        return false;
    }

    exportedUsing<FORMAT>(exporter: ReportExporter<FORMAT>): PromiseLike<FORMAT> {
        return exporter.exportRehearsal(this);
    }

    constructor() {
        super(new DomainEvent(new Rehearsal()));    // todo: DomainEvent? Meh. Can I do better?
        this.parent = undefined;
    }
}

export class Rehearsal {
}

export class ActivityPeriod extends ReportPeriod<Activity> {
    matches(another: Activity): boolean {
        return this.value.equals(another);
    }

    exportedUsing<FORMAT>(exporter: ReportExporter<FORMAT>): PromiseLike<FORMAT> {
        return exporter.exportActivity(this);
    }
}

export class ScenePeriod extends ReportPeriod<Scene> {
    matches(another: Scene): boolean {
        return this.value.equals(another);
    }

    exportedUsing<FORMAT>(exporter: ReportExporter<FORMAT>): PromiseLike<FORMAT> {
        return exporter.exportScene(this);
    }
}

==> serenity-js-1.1.0.tgz: package/src/serenity/reporting/rehearsal_report.ts <==
import {
    Activity,
    ActivityFinished,
    ActivityStarts,
    DomainEvent,
    Outcome,
    Photo,
    PhotoAttempted,
    Scene,
    SceneFinished,
    SceneStarts,
} from '../../serenity/domain';
import { ReportExporter } from './report_exporter';

export class RehearsalReport {
    static from(events: Array<DomainEvent<any>>): RehearsalPeriod {
        let previousNode: ReportPeriod<Scene | Activity>;
        let currentNode:  ReportPeriod<Scene | Activity>;

        return events.reduce((fullReport, event) => {
            switch (event.constructor) {    // tslint:disable-line:switch-default   - ignore other events
                case SceneStarts:
                    currentNode = fullReport.append(new ScenePeriod(event));
                    break;
                case ActivityStarts:
                    currentNode = currentNode.append(new ActivityPeriod(event));
                    break;
                case SceneFinished:
                    currentNode = currentNode.concludeWith(event);
                    break;
                case ActivityFinished:
                    // The Photographer triggers a PhotoAttempted event after the ActivityFinished,
                    // that's why we need to cache both the previous and the current node
                    previousNode = currentNode;
                    currentNode  = currentNode.concludeWith(event);
                    break;
                case PhotoAttempted:
                    [previousNode, currentNode]
                        .filter(node => !! node && node.matches(event.value.activity))
                        .forEach(node => node.attach(event.value.photo));
                    break;
            }

            return fullReport;
        }, new RehearsalPeriod());
    }
}

export abstract class ReportPeriod<T> {
    public parent:    ReportPeriod<any>;
    public value:     T;

    public children:  Array<ReportPeriod<any>> = [];
    public outcome:   Outcome<T>;

    public startedAt:  number;
    protected finishedAt: number;

    private promisedPhotos: Array<PromiseLike<Photo>> = [];

    constructor(start: DomainEvent<T>) {
        this.value     = start.value;
        this.startedAt = start.timestamp;
    }

    abstract matches(finished: T): boolean;

    concludeWith(finished: DomainEvent<Outcome<any>>) {
        this.finishedAt = finished.timestamp;
        this.outcome    = finished.value;

        if (! this.parent) {
            return this;
        }

        return this.matches(finished.value.subject)
            ? this.parent
            : this.parent.concludeWith(finished);
    }

    attach(promisedPhoto: PromiseLike<Photo>) {
        this.promisedPhotos.push(promisedPhoto);
    }

    photos(): PromiseLike<Photo[]> {
        // todo: maybe use some constant rather than a blunt "undefined"?
        return Promise.all(this.promisedPhotos).then(photos => photos.filter(p => p !== undefined));
    }

    duration() {
        return this.finishedAt - this.startedAt;
    }

    append(child: ReportPeriod<Scene | Activity>): ReportPeriod<Scene | Activity> {
        child.parent = this;
        this.children.push(child);

        return child;
    }

    abstract exportedUsing<FORMAT>(exporter: ReportExporter<FORMAT>): PromiseLike<FORMAT>;
}

export class RehearsalPeriod extends ReportPeriod<Rehearsal> {
    matches(finished: Rehearsal): boolean {
        return false;
    }

    exportedUsing<FORMAT>(exporter: ReportExporter<FORMAT>): PromiseLike<FORMAT> {
        return exporter.exportRehearsal(this);
    }

    constructor() {
        super(new DomainEvent(new Rehearsal()));    // todo: DomainEvent? Meh. Can I do better?
        this.parent = undefined;
    }
}

export class Rehearsal {
}

export class ActivityPeriod extends ReportPeriod<Activity> {
    matches(another: Activity): boolean {
        return this.value.equals(another);
    }

    exportedUsing<FORMAT>(exporter: ReportExporter<FORMAT>): PromiseLike<FORMAT> {
        return exporter.exportActivity(this);
    }
}

export class ScenePeriod extends ReportPeriod<Scene> {
    matches(another: Scene): boolean {
        return this.value.equals(another);
    }

    exportedUsing<FORMAT>(exporter: ReportExporter<FORMAT>): PromiseLike<FORMAT> {
        return exporter.exportScene(this);
    }
}
